.transparent {
    background-color: transparent;
}

row.thumbnail {
    border: solid 1px @borders;
    box-shadow: 0px 0px 3px alpha(@borders, 0.8);;
    border-radius: 16px;
}

grid-item.thumbnail {
    padding: 6px;
    font-size: 10px;
}

grid-item image {
    margin: 12px 24px;
    font-size: 10px;
    border: solid 1px @borders;
    border-radius: 4px;
    box-shadow: 0px 0px 4px alpha(@borders, 0.8);
}

grid-item {
    background-color: transparent;
}

grid-item.light {
    color: #000003;
}

grid-item.dark {
    color: #eeeeec;
}

arrow-frame {
    border: solid 1px;
    border-bottom-width: 0px;
    border-top-width: 0px;
    background-color: @theme_bg_color;
    border-color: @borders;
}

arrow-frame:dir(ltr) {
    border-right-width: 0px;
}

arrow-frame:dir(rtl) {
    border-left-width: 0px;
}

label.main-title {
    font-size: 28px;
}


/*
 * GtdTaskRow, GtdNewTaskRow & GtdDndRow
 */

taskrow {
    border-radius: 2px;
    background-color: @theme_base_color;
    box-shadow: 0 0 2px alpha(black, 0.33);
}

taskrow:dir(ltr):not(.new-task-row) { padding-left: 6px;  }
taskrow:dir(rtl):not(.new-task-row) { padding-right: 6px; }

taskrow.priority-low:dir(ltr)    { border-left:  solid 6px #3465a4; padding-left:  0; }
taskrow.priority-low:dir(rtl)    { border-right: solid 6px #3465a4; padding-right: 0; }

taskrow.priority-medium:dir(ltr) { border-left:  solid 6px #f57900; padding-left:  0; }
taskrow.priority-medium:dir(rtl) { border-right: solid 6px #f57900; padding-right: 0; }

taskrow.priority-hight:dir(ltr)  { border-left:  solid 6px #cc0000; padding-left:  0; }
taskrow.priority-hight:dir(rtl)  { border-right: solid 6px #cc0000; padding-right: 0; }

/* task title entry */
taskrow entry                                      { border: none; background: none; }
taskrow.new-task-row entry                         { padding: 0 18px;                }
taskrow:not(.new-task-row) entry                   { border: solid 1px transparent;  }
taskrow:not(.new-task-row) entry:hover:not(:focus) { border: solid 1px @borders;     }

/* line-through complete tasks */
taskrow.complete entry.title { text-decoration-line: line-through; }

taskrow .close-button.active > image { transition: 200ms cubic-bezier(0.25, 0.46, 0.45, 0.94); -gtk-icon-transform: rotate(-0.5turn); }
taskrow .close-button > image { transition: 200ms cubic-bezier(0.25, 0.46, 0.45, 0.94); -gtk-icon-transform: rotate(0turn); }

/* dnd row */
dndrow frame { border: solid 2px; border-radius: 2px;  }

tasklistview.dark  dndrow frame { background: alpha(white, 0.25); border-color: alpha(white, 0.5);}
tasklistview.light dndrow frame { background: alpha(black, 0.25); border-color: alpha(black, 0.5);}

/* extension list */
list.extension-list row {
    border-bottom: solid 1px @borders
}

/* circle button */
button.circle {
    border-radius: 20px;
}

/*
 * Tasklist view
 */
tasklistview.dark emptylistwidget label,
tasklistview.dark emptylistwidget image {
    color: white;
}

tasklistview.light emptylistwidget label,
tasklistview.light emptylistwidget image {
    color: black;
}
